//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2016
//#
//#
//# File Name:  i2c_transmit_model.v
//# Author:     Earth People Technology
//# Date:       July 26, 2016
//# Revision:   A
//#
//# Development: EPT I2C Slave Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			7/26/16 	1			Created
//#
//#									
//#
//#######################################################################
module i2c_transmit_model
 (
	input wire                           clk,
	input wire                           reset,
	output reg                           i2c_data,
	input wire                           i2c_clock,

	output reg                           transmit_complete,
	input wire	[(1<<3)-1:0]             data_in,
	input wire                           data_valid,
	input wire                           active

			);

	parameter   log_width = 3;

	parameter	IDLE                    = 0,
                READ_DATA_IN            = 1,
                INITIATE_IIC_START      = 2,
                LOAD_SHIFT_REG          = 3,
                SHIFT_BIT_OUT           = 4,
                CHECK_FOR_END_OF_WORD   = 5;

	
//=================================================================================
//	Internal Registers
//=================================================================================
	// Holding Registers
	reg      holding_reg_empty;

	//wire					i2c_data_pre;
	reg	[(1<<log_width)-1:0]	holding_reg;
	reg	[(1<<log_width)-1:0]	shift_reg;
	reg	[1:0]				phase;
	reg					i2c_data_idle;
	reg					i2c_data_select;
	reg					last_data_valid;
	
	reg  [7:0]   bit_counter;

	// Transmit complete registers
	reg					transmit_complete_reg;

//=================================================================================
//	Create the I2C output Data
//=================================================================================


	always @(negedge clk or negedge reset)
	begin
		if( !reset )
			i2c_data <= 1'b0;
		else
		   if(i2c_data_select)
			i2c_data <= shift_reg[(1<<log_width)-1];
	end

//=================================================================================
//	Create a shift out complete register
//=================================================================================

	always @(posedge clk or negedge reset)
	begin
		if (!reset)
		begin
				transmit_complete <= 1'b0;
				transmit_complete_reg <= 1'b0;
		end
		else
		begin
			if((bit_counter == 9'h01) & !transmit_complete_reg )
			begin
				transmit_complete <= 1'b1;
				transmit_complete_reg <= 1'b1;
			end
			else if((bit_counter == 9'h00) & transmit_complete_reg )
			begin
				transmit_complete <= 1'b0;
				transmit_complete_reg <= 1'b0;
			end
			else if(!transmit_complete_reg & ( phase == 2'b0 ))
			begin
				transmit_complete <= 1'b0;
				transmit_complete_reg <= 1'b0;
			end
		end
	end


//=================================================================================
//	Create the I2C Bus Phases and shift registers
//=================================================================================

always @(posedge clk or negedge reset)
   if (!reset)
      begin
			i2c_data_idle <= 1'b1;
			i2c_data_select <= 1'b0;
			last_data_valid <= 1'b0;
			holding_reg_empty <= 1'b1;
			bit_counter <= 0;
			phase <= 2'b00;
      end
   else
   begin
         case (phase)
            2'b00: 
			begin
                if (active)
                begin
                   bit_counter <= ((1<<log_width)+1);
                   phase <= 2'b01;
                end
            end
            2'b01: 
			begin
                if (active)
                begin
                      shift_reg <= data_in;
                      holding_reg_empty <= 1'b1;
                      i2c_data_select <= 1'b1;
                      phase <= 2'b10;
				  //end
                end

                if(!i2c_clock)
				begin
                    i2c_data_select <= 1'b1;
                    phase <= 2'b10;
				end
                else
                begin
                    phase <= 2'b01;
                end
            end
            2'b10: 
			begin
                if (i2c_clock)
                begin
                   if (bit_counter != 0)
				   begin
                      if (bit_counter > 1)
                         shift_reg <= {shift_reg[((1<<log_width)-2):0], 1'b0};
                      phase <= 2'b01;
                      i2c_data_select <= 1'b0;
                      bit_counter <= bit_counter - 1;
                   end
				   else
				      phase <= 2'b11;
                end
                else
                begin
                    phase <= 2'b10;
                end
            end
            2'b11: 
			begin
               i2c_data_idle <= 1'b1;
               i2c_data_select <= 1'b0;
               last_data_valid <= 1'b0;
               holding_reg_empty <= 1'b1;
               phase <= 2'b00;
			end

         endcase
           
         if (data_valid & !last_data_valid)
            begin
               holding_reg <= data_in;
               holding_reg_empty <= 1'b0;
            end
         last_data_valid <= data_valid;
   end

endmodule


