//#################################################################
//	Testbench for the Serial UART Communications module, part of the 
//	
//
//	File name:	tb_serial_uart_comms.v
//	Author:		Richard Jolly
//	Date:		11/24/2008
//
//	Copyright 	Smartronix Inc. 2008
//
//################################################################

`timescale 1ns / 100ps

`define SERIAL_CYCLE 100
`define CYCLE 10
`define TEST_10_REPS 10
//`include "tb_define.v"

module tb_serial_uart_comms;

   //----------------------------------------------
   // Parameter Declarations  
   //----------------------------------------------
	parameter	BEGIN_SECTION			= 8'h0;
	parameter	WRITE_CONFIG_SECTION		= 8'h02;
	parameter	READ_CONFIG_SECTION		= 8'h01;
	parameter	WRITE_SECTION			= 8'h04;
	parameter	READ_SECTION			= 8'h03;

//---------------------------------------------------------------------------------------------------------------
//	wire and register declarations
//---------------------------------------------------------------------------------------------------------------
	
	// Clocks
	reg						clk;
	reg						serial_clk;
	reg						reset;

	integer						logfile;
	reg	[7:0]						num_read_cycles;
	integer						num_write_cycles;
	integer						num_config_read_cycles;
	integer						num_config_write_cycles;

	reg						cardbus_reset;
	reg						register_memory;



	reg	[31:0]					next_address;

	reg	[31:0]					write_word;

	//Host bus model start buttons
	reg						start_config_read_cycle;
	reg						start_config_write_cycle;
	reg						start_read_cycle;
	reg						start_write_cycle;
	reg						write_to_cis_register_block ;

	//cis registers
	reg	[7:0]						cis_register_index;

	//Host bus model write registers
	reg	[7:0]						write_value;

	//Compare read register configuration registers
	wire	[7:0]					configuration_address;
	wire	[7:0]					config_compare_data;
	wire	[7:0]					control_compare_data;


	//Read Control Registers
	wire	[7:0]					read_control_register;
	reg	[7:0]					read_control_register_index;

	//read and write register bytes
	reg	[7:0]					host_read_register;
	wire	[7:0]					write_register_value;

	reg							write_to_read_control_register_block;
	reg	[7:0]					write_read_control_word;

	//Write Control Registers
	wire	[7:0]						write_control_register;
	reg	[7:0]						write_control_register_index;
	reg	[7:0]						write_to_write_control_register_block;
	reg							write_control_word;
	reg	[7:0]						host_write_byte ;

	wire	[25:0]					host_bus_model_addr;
	wire	[15:0]					host_bus_model_data;

	//DUT pass registers
	wire						usb_rxf_n;
	wire						usb_txe_n;
	wire						usb_rd_n;
	wire						usb_wr;
	wire						usb_reset_n;
	wire						usb_pwren_n;
	wire						usb_test;
	wire	[7:0]					usb_data;

	// IIC Serial Communications testbench
	wire						physical_line_iic_data;
	wire						physical_line_iic_clock;

	// Serial Communications Host Model signals
	reg	[15:0]				serial_iic_data_in_model;
	wire						serial_iic_ena_out_model;
	reg						serial_iic_rdy_in_model;
	reg	[7:0]					serial_iic_word_count_in_model;

	wire	[15:0]				serial_iic_data_out_model;
	reg						serial_iic_ena_in_model;
	wire						serial_iic_rdy_out_model;

	// Serial Communications IIC Bus
	reg	[15:0]				serial_iic_data_in;
	wire						serial_iic_ena_out;
	reg						serial_iic_rdy_in;
	wire						serial_iic_word_count_in;

	wire	[15:0]				serial_iic_data_out;
	reg						serial_iic_ena_in;
	wire						serial_iic_rdy_out;

	// Sensor Feedback Registers
	wire	[7:0]					sensor_feedback_control;
	wire	[7:0]					sensor_feedback_receive;
	reg	[7:0]					sensor_feedback_status;

	//Internal registers
	reg	[15:0]				host_transmit_word_1;
	reg	[15:0]				host_transmit_word_2;
	reg	[7:0]					host_transmit_num_of_words;
	reg	[15:0]				host_write_word;



  	// Test Bench verification signals
  	reg [15:0]					dut_received_word;
  	reg [15:0]					dut_num_of_words;

	//Test signals
	wire						host_bus_model_test_signal_1;
	reg	[7:0]					temp_hold_reg;

	// Control and Status Registers for the Serial UART Comms
	wire	[7:0]					serial_uart_control_register_model;
	reg	[7:0]					serial_uart_status_register_model;
	

	wire	[7:0]					serial_uart_control_register;
	reg	[7:0]					serial_uart_status_register;

	// Serial UART Comms Registers
	wire						physical_line_uart_txd_from_dut;
	wire						physical_line_uart_rxd_to_dut;
	wire	[15:0]					serial_uart_receive_register;
	reg	[15:0]					serial_uart_send_register;
	wire	[15:0]					serial_uart_receive_register_model;
	reg	[15:0]					serial_uart_send_register_model ;

//--------------------------------------------------------------------------------------------------------------
//	Assignments
//--------------------------------------------------------------------------------------------------------------
	
//	assign	serial_uart_receive_register	=	1'b0;

//--------------------------------------------------------------------------------------------------------------
//	Generate Internal Clock with 'cycle/2 period
//--------------------------------------------------------------------------------------------------------------

	initial 
	begin
		clk = 0;
		forever
		//begin
			# `CYCLE clk = !clk;
		//end
	end


	initial 
	begin
		serial_clk = 0;
		forever
		//begin
			# `SERIAL_CYCLE serial_clk = !serial_clk;
		//end
	end

//-------------------------------------------------------------------------------------------------------------
//	Apply Stimulus
//-------------------------------------------------------------------------------------------------------------

	initial
	begin


		call_title(BEGIN_SECTION);
		clk						=	1'b0;
		reset						=	1'b1;
		host_transmit_word_1			=	16'h0b;
		host_transmit_word_2			=	16'h88;
		host_transmit_num_of_words		=	8'h2;
		serial_iic_rdy_in_model 		= 	1'b0;
		serial_iic_ena_in 			= 	1'b0;
		num_read_cycles 				=	8'h0;
		host_write_word 				=	16'h0;
		serial_uart_status_register		=	8'h0;
		serial_uart_status_register_model	=	8'h0;


		#(100 * `CYCLE)
		reset	=			1'b0;

		#(50 * `CYCLE)
		reset	=			1'b1;




		repeat(`TEST_10_REPS)
		begin
		//write from the Host Model to DUT
		call_title(WRITE_SECTION);
		repeat(1)
		begin

			host_write_word = host_transmit_word_1;
			#(15000 * `CYCLE )
			write_outgoing_message_serial_uart_comms_model( host_transmit_word_1, host_transmit_num_of_words);


		

		end // repeat(num_read_cycles)


		//Read two words from the DUT 
		call_title(READ_SECTION);
		repeat(1)
		begin


			#( 1000 * `CYCLE )
			read_incoming_message_serial_uart_comms( dut_received_word, num_read_cycles );


			#( 1000 * `CYCLE )
			compare_host_read( host_write_word, dut_received_word, num_read_cycles );
		
			num_read_cycles = num_read_cycles + 1;
			host_transmit_word_1 = host_transmit_word_1 + num_read_cycles;
		end // repeat(num_read_cycles)
		
		#(1000 * `CYCLE);
		end //repeat()


		repeat(`TEST_10_REPS)
		begin
		//write from the DUT to Host Model
		call_title(WRITE_SECTION);
		repeat(1)
		begin
			host_write_word = host_transmit_word_1;

			#(15000 * `CYCLE )
			write_outgoing_message_serial_uart_comms( host_transmit_word_1, host_transmit_num_of_words);


		

		end // repeat(num_read_cycles)


		//Read two words from the DUT 
		call_title(READ_SECTION);
		repeat(1)
		begin


			#( 1000 * `CYCLE )
			read_incoming_message_serial_uart_comms_model( dut_received_word, num_read_cycles );


			#( 1000 * `CYCLE )
			compare_host_read( host_write_word, dut_received_word, num_read_cycles );
		
			num_read_cycles = num_read_cycles + 1;
			host_transmit_word_1 = host_transmit_word_1 + num_read_cycles;
		end // repeat(num_read_cycles)
		
		#(1000 * `CYCLE);
		end //repeat()

	  #(20000 * `CYCLE);
	stopsim;


	end	//stimulus
//--------------------------------------------------
//	Task: call title
//--------------------------------------------------
   task call_title;
	input [7:0] select_title;
      begin
		case ( select_title ) 
		8'h0:
		$display("\n\n\n\nThis testbench will test the Serial UART Comms.v .\n");
		8'h01:
		$display("\n\n\n\nRead Configuration Section of Testbench starting with address \n");
		8'h02:
		$display("\n\n\n\nWrite Configuration Section of Testbench starting with address \n");
		8'h03:
		begin
		$display("\n\n\n\nRead Section of Testbench performing read....\n");
		end
		8'h04:
		$display("\n\n\n\nWrite Section of Testbench \n");
		default:
		$display("\n\n\n\nTest bench running....\n");
		endcase

      end
   endtask // host_bus_read


//--------------------------------------------------
//	Task: read_incoming_message_serial_uart_comms_model
//	For Host Model Only
//	Once the serial_iic_comms receives a message,
//	read available words.
//--------------------------------------------------
   task	read_incoming_message_serial_uart_comms_model;
      output	[15:0]	received_data_in;
	input				read_received_word;
      begin
	
	if( serial_uart_control_register_model[3] )
	begin
		received_data_in = serial_uart_receive_register_model;
		 #(20 * `CYCLE);

		serial_uart_status_register_model[3] = 1'b1;
		 #(20 * `CYCLE);

		serial_uart_status_register_model[3] = 1'b0;
		#(20 * `CYCLE);

	end
	else 
		$display("Failded to read Receive UART from DUT. \n
				serial_uart_control_register[3] .\n");

      end
   endtask 

//--------------------------------------------------
//	Task: write_outgoing_message_serial_uart_comms_model
//	For Host Model only
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	write_outgoing_message_serial_uart_comms_model;
      input		[15:0]	transmit_data_out_1;
	input		[7:0]		num_of_words;

	reg				first_time_thru_loop;



      begin

	first_time_thru_loop = 1'b0;

	while( num_of_words ) 
	begin

			serial_uart_send_register_model = transmit_data_out_1;
	 		#(200 * `CYCLE);

			serial_uart_status_register_model[1] = 1'b1;
	 		#(200 * `CYCLE);

//			wait( serial_uart_control_register_model[2] );
//			#(200 * `CYCLE);

			serial_uart_status_register_model[1] = 1'b0;
	 		#(200 * `CYCLE);

			num_of_words = num_of_words - 1;
			first_time_thru_loop <= first_time_thru_loop + 1;
			#(200 * `CYCLE);

	end

      end
   endtask 

//--------------------------------------------------
//	Task: read_incoming_message_serial_uart_comms
//	For Host Model Only
//	Once the serial_iic_comms receives a message,
//	read available words.
//--------------------------------------------------
   task	read_incoming_message_serial_uart_comms;
      output	[15:0]	received_data_in;
	input				read_received_word;
      begin
	
	if( serial_uart_control_register[3] )
	begin
		received_data_in = serial_uart_receive_register;
		 #(20 * `CYCLE);

		serial_uart_status_register[3] = 1'b1;
		 #(20 * `CYCLE);

		serial_uart_status_register[3] = 1'b0;
		#(20 * `CYCLE);

	end
	else 
		$display("Failded to read Receive UART from DUT. \n
				serial_uart_control_register[3] .\n");

      end
   endtask 

//--------------------------------------------------
//	Task: write_outgoing_message_serial_uart_comms
//	For Host Model only
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	write_outgoing_message_serial_uart_comms;
      input		[15:0]	transmit_data_out_1;
	input		[7:0]		num_of_words;

	reg				first_time_thru_loop;



      begin

	first_time_thru_loop = 1'b0;

	while( num_of_words ) 
	begin

			serial_uart_send_register = transmit_data_out_1;
	 		#(200 * `CYCLE);

			serial_uart_status_register[1] = 1'b1;
	 		#(200 * `CYCLE);

//			wait( serial_uart_control_register[2] );
//			#(200 * `CYCLE);

			serial_uart_status_register[1] = 1'b0;
	 		#(200 * `CYCLE);

			num_of_words = num_of_words - 1;
			first_time_thru_loop <= first_time_thru_loop + 1;
			#(200 * `CYCLE);

	end

      end
   endtask 



//--------------------------------------------------
//	Task: read_incoming_message_host_model
//	For Host Model Only
//	Once the serial_iic_comms receives a message,
//	read available words.
//--------------------------------------------------
   task	read_incoming_message_host_model;
      output	[15:0]	received_data_in;
      output	[15:0]	received_num_of_words;
	input				read_received_word;
      begin
	
	if( serial_iic_rdy_out_model)
	begin
		serial_iic_ena_in_model = 1'b1;
		 #(20 * `CYCLE);

		received_data_in = serial_iic_data_in_model;
		 #(20 * `CYCLE);

		serial_iic_ena_in_model= 1'b0;
		#(20 * `CYCLE);

	end
	else 
		$display("Failded to read Receive FIFO from Host Model. \n
				serial_iic_rdy_out_model not asserted.\n");

      end
   endtask //  host_bus_read

//--------------------------------------------------
//	Task: write_outgoing_message_host_model
//	For Host Model only
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	write_outgoing_message_host_model;
      input		[15:0]	transmit_data_out_1;
      input		[15:0]	transmit_data_out_2;
	input		[7:0]		num_of_words;

	reg				first_time_thru_loop;



      begin

	first_time_thru_loop = 1'b0;

	while( num_of_words ) 
	begin
		serial_iic_word_count_in_model = num_of_words;

			if( !first_time_thru_loop )
				serial_iic_data_in_model = transmit_data_out_1;
			else
				serial_iic_data_in_model = transmit_data_out_2;
			 #(200 * `CYCLE);

			serial_iic_rdy_in_model = 1'b1;
	 		#(200 * `CYCLE);

			wait( serial_iic_ena_out_model );//serial_iic_ena_out_model
			#(200 * `CYCLE);

			serial_iic_rdy_in_model = 1'b0;
	 		#(200 * `CYCLE);

			num_of_words = num_of_words - 1;
			first_time_thru_loop <= first_time_thru_loop + 1;
			#(200 * `CYCLE);

	end

      end
   endtask //  host_bus_write

//--------------------------------------------------
//	Task: serial_iic_read_incoming_message
//	Once the serial_iic_comms receives a message,
//	read available words.
//--------------------------------------------------
   task	serial_iic_read_incoming_message;
      output	[15:0]	received_data_in;
      output	[15:0]	received_num_of_words;
	input		[7:0]		read_cycles;
      begin
	
	if( serial_iic_rdy_out)
	begin
		serial_iic_ena_in = 1'b1;
		 #(20 * `CYCLE);

		received_data_in = serial_iic_data_out;
		 #(20 * `CYCLE);
	
		wait( !serial_iic_rdy_out );
		serial_iic_ena_in = 1'b0;
		#(20 * `CYCLE);

	end
	else 
		$display("Failded to read Receive FIFO from DUT. \n
		serial_iic_rdy_out not asserted. Read cycle %h\n",read_cycles);

      end
   endtask //  host_bus_read

//--------------------------------------------------
//	Task: serial_iic_write_outgoing_message
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	serial_iic_write_outgoing_message;
      input		[15:0]	transmit_data_out_1;
      input		[15:0]	transmit_data_out_2;
	input				num_of_words;
      begin

	
	while( num_of_words ) 
	begin
		if( !serial_iic_ena_out )
		begin
			serial_iic_data_in = transmit_data_out_1;
			 #(20 * `CYCLE);

			serial_iic_rdy_in_model = 1'b1;
	 		#(20 * `CYCLE);

			wait( !serial_iic_ena_out );
			#(20 * `CYCLE);

			num_of_words = num_of_words - 1;
			#(20* `CYCLE);
		end

	end

      end
   endtask //  host_bus_write


//--------------------------------------------------
//	Task: read_incoming_message
//	Once the serial_iic_comms has stored a message,
//	read available words.
//--------------------------------------------------
   task	read_incoming_message;
      output	[15:0]	received_data_in;
	input		[7:0]		read_cycles;

	reg				sensor_feedback_ena_in;
	reg	[31:0]		error_count;

      begin
	sensor_feedback_ena_in = 1'b0;
	error_count	= 32'h0;
	
	wait( sensor_feedback_control[2] && 1'b1 )
	begin
		if( error_count <= 32'h01000000 )
			error_count = error_count + 1;
		else 
			$display("Failded to read Receive FIFO from DUT. \n
			serial_iic_rdy_out not asserted. Read cycle %h\n",read_cycles);
	end

		sensor_feedback_ena_in = 1'b1;
		 #(20 * `CYCLE);

		received_data_in = sensor_feedback_receive;
		 #(200 * `CYCLE);
	
		sensor_feedback_status[0] = 1'b1;
		wait( !sensor_feedback_control[2] && 1'b1 );
		sensor_feedback_ena_in  = 1'b0;
		#(200 * `CYCLE);
		sensor_feedback_status[0] = 1'b0;



      end
   endtask //  host_bus_read

//--------------------------------------------------
//	Task: write_outgoing_message
//	Write two words to the serial_iic_comms,
//	when done, start the transmit process.
//--------------------------------------------------
   task	write_outgoing_message;
      input		[15:0]	transmit_data_out_1;
      input		[15:0]	transmit_data_out_2;
	input				num_of_words;
      begin

	
	while( num_of_words ) 
	begin
		if( !serial_iic_ena_out )
		begin
			serial_iic_data_in = transmit_data_out_1;
			 #(20 * `CYCLE);

			serial_iic_rdy_in_model = 1'b1;
	 		#(20 * `CYCLE);

			wait( !serial_iic_ena_out );
			#(20 * `CYCLE);

			num_of_words = num_of_words - 1;
			#(20* `CYCLE);
		end

	end

      end
   endtask //  host_bus_write


//--------------------------------------------------
//	Task: Configuration read register memory
//--------------------------------------------------
   task host_bus_config_read;
      input [31:0] read_register_address;
      begin
	next_address = read_register_address;
	start_config_read_cycle = 1'b1;
	 #(2 * `CYCLE);
	start_config_read_cycle = 1'b0;
	 #(2 * `CYCLE);
      end
   endtask // host_bus_config_read


//--------------------------------------------------
//	Task: Configuration write register memory
//--------------------------------------------------
   task host_bus_config_write;
      input [31:0] write_register_address;
      input [31:0] write_register_value;
      begin
	next_address = write_register_address;
	write_value = write_register_value;
	start_config_write_cycle = 1'b1;
	 #(2 * `CYCLE);
	start_config_write_cycle = 1'b0;
	 #(2 * `CYCLE);
      end
   endtask // host_bus_config_write

//--------------------------------------------------
//	Task: read register memory
//--------------------------------------------------
   task host_bus_read;
      input [7:0] read_register_address;
       input [7:0] read_register_value;
     begin
	next_address = read_register_address;
	host_read_register = read_register_value;
	start_read_cycle = 1'b1;
	 #(2 * `CYCLE);
	start_read_cycle = 1'b0;
	 #(2 * `CYCLE);
      end
   endtask // host_bus_read


//--------------------------------------------------
//	Task: write register memory
//--------------------------------------------------
   task host_bus_write;
      input [7:0] write_register_address;
       input [7:0] write_register_value;
      begin
	next_address = write_register_address;
	start_write_cycle = 1'b1;
	 #(2 * `CYCLE);
	start_write_cycle = 1'b0;
	 #(2 * `CYCLE);
      end
   endtask // host_bus_write

//--------------------------------------------------
//	Task: compare read register 
//--------------------------------------------------
   task compare_host_read;
	input [31:0]	write_register;
	input [31:0]	read_register;
	input	[7:0]		read_cycle_number;
	reg	start_compare_read_cycle;


	begin
		start_compare_read_cycle = 1'b0;


		$display("Performing the Compare of read and write words\n");
		start_compare_read_cycle = 1'b1;
	 	#(200* `CYCLE);
		if( write_register[15:0] != read_register[15:0] )
			$display("Compare failed at read cycle %h\n", read_cycle_number);
		else
			$display("Compare Succeeded!!! at read cycle %h\n", read_cycle_number);
		start_compare_read_cycle = 1'b0;
		 #(200 * `CYCLE);
		
	end
    endtask

//--------------------------------------------------
//	Task: test_bench_config_write
//	update the configuration block with new word
//--------------------------------------------------
   task	test_bench_config_write;
      input [31:0] write_tb_address;
      input [31:0] write_tb_value;
      begin
	cis_register_index = write_tb_address;
	write_word = write_tb_value;
	write_to_cis_register_block = 1'b1;
	 #(2 * `CYCLE);
	write_to_cis_register_block = 1'b0;
	 #(2 * `CYCLE);
      end
   endtask // host_bus_read

//--------------------------------------------------
//	Task: test_bench_write
//	update the configuration block with new word
//--------------------------------------------------
   task	test_bench_write;
      input [31:0] write_tb_address;
      input [31:0] write_tb_value;
      begin
	write_control_register_index = write_tb_address;
	 write_control_word = write_tb_value;
	write_to_write_control_register_block = 1'b1;
	 #(2 * `CYCLE);
	write_to_write_control_register_block = 1'b0;
	 #(2 * `CYCLE);
      end
   endtask //  host_bus_read

   //-----------------------------------------------
   // Task: stopsim
   //-----------------------------------------------
   task stopsim;
      begin
   	 $fdisplay(logfile,"Simulation Finished at time %0t\n",$time);
	 $stop;
	 $fclose(logfile);
      end
   endtask // stopsim


  
   //-----------------------------------------------
   // Instantiate Memory Test DUT
   //-----------------------------------------------
   serial_uart_comms						DUT
    (
		.CLK							(clk),
		.RESET						(reset),

		.UART_RXD_IN					(physical_line_uart_txd_from_dut),
		.UART_TXD_OUT					(physical_line_uart_rxd_to_dut),

		.CONTROL_REGISTER					(serial_uart_control_register ),
		.STATUS_REGISTER					(serial_uart_status_register),
		.RECEIVE_REGISTER					(serial_uart_receive_register),
		.SEND_REGISTER					(serial_uart_send_register)
	
	
     );

   //-----------------------------------------------
   // Instantiate Host Bus Model
   //-----------------------------------------------
   serial_uart_comms_model 					serial_uart_comms_model_inst
     (
		.CLK							(clk),
		.RESET						(reset),

		.UART_RXD_IN					(physical_line_uart_rxd_to_dut),
		.UART_TXD_OUT					(physical_line_uart_txd_from_dut),

		.CONTROL_REGISTER					(serial_uart_control_register_model),
		.STATUS_REGISTER					(serial_uart_status_register_model),
		.RECEIVE_REGISTER					(serial_uart_receive_register_model), 
		.SEND_REGISTER					(serial_uart_send_register_model )
	
 
      );   


endmodule // tb_serial_uart_comms.v


