/******************************************************
 * $Workfile::                                       
 * $Author::                                         
 * $Date::                                           
 * $Revision::                                       
 *
 *
 * This header files defines the input/output
******************************************************/
`include "../src/define.v"
`include "../src/tb_define.v"

`timescale 1 ns / 10 ps

module uart_tiny_slave
  (
	input wire                 CLK,
	input wire                 RST_N,

  
	output wire                 UART_OUT, //To external 
    input wire                 UART_IN, //From external
  
	input wire                 USB_RXF_N,
	input wire                 USB_TXE_N,
    inout wire  [7:0]          USB_DATA,

	output wire                USB_RD_N,
	output reg                 USB_WR,
	output reg                 USB_SI,

	output wire                TR_DIR_1,
	output wire                TR_OE_1,
	output wire                TR_DIR_2,
	output wire                TR_OE_2,
	output wire                TR_DIR_3,
	output wire                TR_OE_3

	);



   //----------------------------------------------
   // Parameter Declarations                        
   //----------------------------------------------
   parameter IDLE                   = 0,
	     RECEIVE_DATA				= 1,
		 STORE_BYTE		    	    = 2,
		 SEND_TO_HOST			    = 3,
		 RECEIVE_COMPLETE			= 4,
		 WRITE_REGISTER 		    = 5,
		 LOAD_SHIFT_REG				= 6,
		 DELAY_WRITE_DATA_BYTE      = 7,
	     WRITE_DATA_BYTE			= 8,
	     WRITE_DATA_COMPLETE		= 9;
		 
`ifdef FAST_UART
		parameter COUNT_TO_DECIMAL       = 4;
		parameter COUNT_TO_DECIMAL_3     = 8'h36;
		parameter COUNT_TO_DECIMAL_2     = 8'h23; // 66 MHz
		parameter COUNT_TO_DECIMAL_1     = 5;
`else
//		parameter COUNT_TO_DECIMAL       = 54;
		parameter COUNT_TO_DECIMAL_4     = 8'h6d;
		parameter COUNT_TO_DECIMAL_3     = 8'h36;
		parameter COUNT_TO_DECIMAL_2     = 8'h23; // 66 MHz
		parameter COUNT_TO_DECIMAL_1     = 5;
		parameter COUNT_TO_DECIMAL       = 4;
`endif

   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
   	reg [11:0]                 state, next;

    // Bit Count Registers
	//reg [3:0]                  in_bit_count;
	//reg [3:0]                  out_bit_count;
	
	
	// Storage Registers
	reg [7:0]                  rcv_storage_reg;
	reg [2:0]                  store_reg_state; 
	
	
	//Metastable Registers
	reg                        uart_in_reg;
	reg                        usb_rxf_reg;
	
	//write_tx_byte_reg for signalling the UART transmitter
	//to start transmitting
	reg                        write_tx_byte_reg;

	//STATE_OUT Regisetrs
    //reg [3:0]                  STATE_OUT;

	//UART Interface signals
	wire [7:0]                  rx_data;
    wire                       rx_byte_present;
    wire                       UART_SERIAL_RX;
    wire                       en_16_x_baud;
    reg                        rx_byte_read;
	wire [7:0]                  tx_data_reg;
    wire                       tx_empty;
    reg                        write_tx_byte;
    //wire                       UART_SERIAL_TX;
	
	//Automated UART write
    reg                        automated_write;
    reg  [7:0]                 automated_write_byte;
    reg  [25:0]                automated_write_counter;
	
   	
`ifdef SIM
   reg [8*26:1] state_name;
`endif

   //------------------------------------------------
   // Signal Assignments  
   //------------------------------------------------
   assign USB_DATA = state[SEND_TO_HOST]   ? rcv_storage_reg : 8'hz;
   assign USB_RD_N = state[WRITE_REGISTER] ? 1'b0 : 1'b1;
   assign tx_data_reg = state[SEND_TO_HOST] ? 8'hz : (automated_write ? automated_write_byte : USB_DATA);
   //assign write_tx_byte = state[WRITE_DATA_BYTE] ? 1'b0: 1'b1;

`ifdef FAST_UART
	reg	[7:0]			counter_reg;
//	assign 	 clk16x = (counter_reg == (COUNT_TO_DECIMAL_3 - 1) );
	assign 	 clk16x = (counter_reg == (COUNT_TO_DECIMAL_2 - 1) );
`else
	reg	[7:0]			counter_reg;
//	assign 	 clk16x = (counter_reg == (COUNT_TO_DECIMAL_3 - 1) );
	assign 	 clk16x = (counter_reg == (COUNT_TO_DECIMAL_2 - 1) );
`endif
	
   assign TR_DIR_1  = 1'b1;
   assign TR_OE_1  = 1'b0;
   assign TR_DIR_2  = 1'b1;
   assign TR_OE_2  = 1'b0;
   //assign TR_DIR_3  = 1'b1;
   assign TR_DIR_3  = state[WRITE_DATA_COMPLETE] ? 1'b0 : 1'b1;
   assign TR_OE_3  = 1'b0;


	
   //-----------------------------------------------
   // Create a State Test Output Signal
   //-----------------------------------------------
/*   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  STATE_OUT <= 0;
	else
	  begin
	     if (state[IDLE])
           STATE_OUT <= 4'h0;
	     if (state[RECEIVE_DATA])
	       STATE_OUT <= 4'h1;
	     if (state[SEND_TO_HOST])
	       STATE_OUT <= 4'h2;
	     if (state[RECEIVE_COMPLETE])
	       STATE_OUT <= 4'h3;
	     if (state[WRITE_REGISTER])
	       STATE_OUT <= 4'h4;
	     if (state[LOAD_SHIFT_REG])
	       STATE_OUT <= 4'h5;
	     if (state[DELAY_WRITE_DATA_BYTE])
	       STATE_OUT <= 4'h6;
	     if (state[WRITE_DATA_BYTE])
	       STATE_OUT <= 4'h7;
	     if (state[WRITE_DATA_COMPLETE])
	       STATE_OUT <= 4'h8;
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)
*/

   //-----------------------------------------------
   // Automated write count (byte) to the 
   // the UART receive module. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	     automated_write <= 1'b0;
	     automated_write_byte <= 0;
		 automated_write_counter <= 26'h0000000;
	end
	else
	  begin
/*	     automated_write_counter <= automated_write_counter + 1'd1;
	     case (automated_write_counter)
		 `START_AUTO_WRITE:
	         automated_write <= 1'b1;
		 `AUTO_WRITE_IDLE:
	         automated_write <= 1'b0;
		 `AUTO_WRITE_RESET_COUNT:
		 begin
*/		     automated_write_counter <= 0;
/*			 automated_write_byte <= automated_write_byte + 1'd1;
		 end
		 endcase
*/	  end
     end // always @ (posedge CLK or negedge RST_N)

//==================================================================
// Transmit: write_tx_byte Signal 
//==================================================================
	always @(posedge CLK or negedge RST_N)
	begin
		if (!RST_N)
		begin
			write_tx_byte <= 1'b1;
			write_tx_byte_reg <= 1'b0;
		end
		else 
		begin
			if(state[WRITE_DATA_BYTE] && !write_tx_byte_reg )
			begin
				write_tx_byte <= 1'b0;
				write_tx_byte_reg <= 1'b1;
			end
			else if(state[WRITE_DATA_BYTE] && write_tx_byte_reg )
			begin
				write_tx_byte <= 1'b1;
				write_tx_byte_reg <= 1'b1;
			end
			else if( state[WRITE_DATA_COMPLETE] )
			begin
				write_tx_byte <= 1'b1;
				write_tx_byte_reg <= 1'b0;
		    end
		end	
	end // always @ (posedge CLK or negedge RESET)   
   
   //-----------------------------------------------
   // rx_byte_read register reads the available byte
   // in the UART receive module. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	     rx_byte_read <= 1'b1;
		 rcv_storage_reg <= 8'he5;
	end
	else
	  begin
	     if (state[STORE_BYTE])
		 begin
	         rx_byte_read <= 1'b0;
			 rcv_storage_reg <= rx_data;
		 end
		 else
		 begin
	         rx_byte_read <= 1'b1;
			 rcv_storage_reg <= rcv_storage_reg;
		 end
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   //Register the USB_RXF_N input signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     usb_rxf_reg <= 1;
	else
	  begin
	     if (!USB_RXF_N)
	         usb_rxf_reg <= 0;
		 else
	         usb_rxf_reg <= 1;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // USB_WR output register based on state[RECEIVE_COMPLETE], 
   // !USB_TXE_N. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     USB_WR <= 1;
	else
	  begin
	     if (state[SEND_TO_HOST] && !USB_TXE_N )
	         USB_WR <= 0;
		 else
	         USB_WR <= 1;
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Register the UART_IN input with uart_in_reg. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	 uart_in_reg <= 1'b1;
	else
	  begin
	     if ( UART_IN )
	       uart_in_reg <= 1'b1;
	     else 
	       uart_in_reg <= 1'b0;
	  end
     end // always @ (posedge CLK or negedge RST_N)

	  
   //--------------------------------------------------
   // UART State Machine
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state <= 12'h00;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end

   always @ ( /*AUTOSENSE*/state or usb_rxf_reg or tx_empty
		 or uart_in_reg or store_reg_state or rx_byte_present
		 or USB_DATA or USB_RD_N or USB_RXF_N or USB_TXE_N
		 or UART_OUT or automated_write_counter)
     begin
	next = 12'h00;

	if (state[IDLE])
	begin
        if ( !uart_in_reg )
			next[RECEIVE_DATA] = 1'b1;
		else if(!usb_rxf_reg)
			next[WRITE_REGISTER] = 1'b1;
		else if(automated_write)
			next[LOAD_SHIFT_REG] = 1'b1;
		else
			next[IDLE] = 1'b1;
	end

	if (state[RECEIVE_DATA])
	  begin
	     if (rx_byte_present)
	       next[STORE_BYTE] = 1'b1;
	     else
	       next[RECEIVE_DATA] = 1'b1;
	  end

	  if (state[STORE_BYTE]) 
	  begin
	     if (!rx_byte_present)
	       next[SEND_TO_HOST] = 1'b1;
	     else
	         next[STORE_BYTE] = 1'b1;
	  end 	     

	if (state[SEND_TO_HOST]) 
	  begin
	     if (USB_TXE_N)
			  next[RECEIVE_COMPLETE] = 1'b1;
	     else
	         next[SEND_TO_HOST] = 1'b1;
	  end 	     

	  if (state[RECEIVE_COMPLETE]) 
	  begin
	     if (uart_in_reg)
			  next[IDLE] = 1'b1;
	     else
	         next[RECEIVE_COMPLETE] = 1'b1;
	  end 	     

	  if (state[WRITE_REGISTER])
	  begin
	     if (!USB_RD_N)
	       next[LOAD_SHIFT_REG] = 1'b1;
	     else 
	       next[WRITE_REGISTER] = 1'b1;
	  end

	if (state[LOAD_SHIFT_REG])
	       next[DELAY_WRITE_DATA_BYTE] = 1'b1;

	if (state[DELAY_WRITE_DATA_BYTE])
	  begin
	       next[WRITE_DATA_BYTE] = 1'b1;
	  end

	if (state[WRITE_DATA_BYTE])
	  begin
	     if (!tx_empty)
	       next[WRITE_DATA_COMPLETE] = 1'b1;
	     else
	       next[WRITE_DATA_BYTE] = 1'b1;
	  end

	if (state[WRITE_DATA_COMPLETE])
	  begin
	     if ( tx_empty )
	     //if ( UART_OUT )
	       next[IDLE] = 1'b1;
	     else
	       next[WRITE_DATA_COMPLETE] = 1'b1;
	  end

	
`ifdef SIM
	if (state == (1 << IDLE))
	  state_name = "IDLE";
	else if (state == (1 << RECEIVE_DATA))
	  state_name = "RECEIVE_DATA";
	else if (state == (1 << SEND_TO_HOST))
	  state_name = "SEND_TO_HOST";
	else if (state == (1 << RECEIVE_COMPLETE))
	  state_name = "RECEIVE_COMPLETE";
	else if (state == (1 << STORE_BYTE))
	  state_name = "STORE_BYTE";
	else if (state == (1 << WRITE_REGISTER))
	  state_name = "WRITE_REGISTER";
	else if (state == (1 << LOAD_SHIFT_REG))
	  state_name = "LOAD_SHIFT_REG";
	else if (state == (1 << DELAY_WRITE_DATA_BYTE))
	  state_name = "DELAY_WRITE_DATA_BYTE";
	else if (state == (1 << WRITE_DATA_BYTE))
	  state_name = "WRITE_DATA_BYTE";
	else if (state == (1 << WRITE_DATA_COMPLETE))
	  state_name = "WRITE_DATA_COMPLETE";
`endif
     end // always @ (...
   
//==============================================================================
//	Instantiate the 16x Clock Counter
//==============================================================================
`ifdef FAST_UART
   // For Simulation use a much faster clock to speed up test
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
 	  counter_reg <= 0;
 	else
 	  begin
 	     if ( counter_reg == COUNT_TO_DECIMAL_2 )
 	       counter_reg <= 0;
 	     else
 	       counter_reg <= counter_reg + 1;
 	  end
     end // always @ (posedge CLK or negedge RST_N)
`else
   // For Simulation use a much faster clock to speed up test
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
 	  counter_reg <= 0;
 	else
 	  begin
 	     if ( counter_reg == COUNT_TO_DECIMAL_2 )
 	       counter_reg <= 0;
 	     else
 	       counter_reg <= counter_reg + 1;
 	  end
     end // always @ (posedge CLK or negedge RST_N)
`endif

   //----------------------------------------------
   // Instantiate the UART_top Leaf FSM Modules
   //----------------------------------------------
   uart_top uart_inst
     (
      .CLK                    (CLK),
      .RST_N                  (RST_N),
      .DOUT                   (rx_data),
      .DATA_READY             (rx_byte_present),
      .RXD                    (UART_IN),
      .CLK16X                 (clk16x),
      .RDN                    (rx_byte_read),
      .DIN                    (tx_data_reg),
      .TX_BUFFER_EMPTY        (tx_empty),
      .WRN                    (write_tx_byte),
      .SDO                    (UART_OUT)
      );

endmodule // uart_slave

